#define F_CPU	18432000L

#include <avr/io.h>
#include <avr/interrupt.h>
#include "uart.h"

#define LP_PCINT 0xc1
#define LP_END	 0xcf

/* has to be a power of two. beware of it's size, so stack won't scribble this data */
#define times_size 0x10

/* circular buffer */
/* gcc sucks with structs */
static volatile uint8_t data[times_size * 4];

/* to-be-reported/written position */
register uint8_t timex asm("r2");
register uint8_t overflow asm("r3");
register uint8_t rpos asm("r4");
register uint8_t wpos asm("r5");

ISR(PCINT_vect)
{
	uint8_t timel = TCNT1L;
	uint8_t timeh = TCNT1H;
	if ((wpos+1) % times_size != rpos && overflow == 0) {
		volatile uint8_t *addr = data + wpos*4;
		*addr++ = timel;
		*addr++ = timeh;
		*addr++ = timex;
		*addr++ = PINB;

		wpos = (wpos + 1) % times_size;
	} else {
		overflow = 1;
	}
}

ISR(TIMER1_OVF_vect)
{
	timex++;
}

void probe_pcint()
{
	/* no prescaling */
	TCCR1B = _BV(CS10);

	/* enable PCINT0..7 inputs */
	PORTB = 0;
	DDRB = 0;

	/* enable PCINT */
	GIMSK |= _BV(PCIE);
	/* enable some pins */
	PCMSK = uart_read();
	/* enable timer overflow int */
	TIMSK |= _BV(TOIE1);

	sei();
	while (1) {
		/* get out */
		if (uart_read_ready()) {
			if (uart_read() == LP_END)
				break;
		}

		if (rpos != wpos) {
			volatile uint8_t *addr = data + rpos*4;
			uart_write(*addr++);
			uart_write(*addr++);
			uart_write(*addr++);
			uart_write(*addr++);

			rpos = (rpos + 1) % times_size;
		} else {
			/* empty buffer */
			if (overflow) {
				uart_write(0xff);
				uart_write(0xff);
				uart_write(0xff);
				uart_write(0xff);
			}
			overflow = 0;
		}
	}
}

int main()
{
	uart_init(115200);
	for (;;) {
		uint8_t c;
		c = uart_read();
		switch (c) {
		case LP_PCINT:
			probe_pcint();
			break;
		default:
			uart_write(c);
		}
	}
	return 0;
}
